/*
 *   $Id: ssvn_system.c,v 1.5 2025/03/24 04:13:11 snw Exp $
 *    ^$SYSTEM ssvn
 *
 *  
 *   Author: Serena Willis <snw@coherent-logic.com>
 *    Copyright (C) 1998 MUG Deutschland
 *    Copyright (C) 2020, 2025 Coherent Logic Development LLC
 *
 *
 *   This file is part of FreeM.
 *
 *   FreeM is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Affero Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   FreeM is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Affero Public License for more details.
 *
 *   You should have received a copy of the GNU Affero Public License
 *   along with FreeM.  If not, see <https://www.gnu.org/licenses/>.
 *
 *   $Log: ssvn_system.c,v $
 *   Revision 1.5  2025/03/24 04:13:11  snw
 *   Replace action macro dat with fra_dat to avoid symbol conflict on OS/2
 *
 *   Revision 1.4  2025/03/09 19:50:47  snw
 *   Second phase of REUSE compliance and header reformat
 *
 *
 * SPDX-FileCopyrightText:  (C) 2025 Coherent Logic Development LLC
 * SPDX-License-Identifier: AGPL-3.0-or-later
 **/

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if defined(HAVE_STDINT_H)
# include <stdint.h>
#endif
#include "mref.h"
#include "mpsdef.h"
#include "shmmgr.h"
#include "iniconf.h"

void ssvn_system_set(char *sub)
{
    char k_buf[255];
    char d_buf[255];

    d_buf[0] = '\201';
    d_buf[1] = '\201';
    snprintf (k_buf, 254, "^$SYSTEM\202%s\201", sub);
    
    symtab_shm (set_sym, k_buf, d_buf);
    
}

void ssvn_system_update(void)
{
    FILE *fp;
    char line[255];
    char line_ns[255];

    char k_buf[255];
    char d_buf[255];

    ssvn_system_set ("MAPPINGS");
    ssvn_system_set ("MAINTENANCE_MODE");
    ssvn_system_set ("DAEMON_PID");
    ssvn_system_set ("MWAPI");
    ssvn_system_set ("EVENT");
    ssvn_system_set ("ERRMSG");
    ssvn_system_set ("NAME_LENGTH");
    ssvn_system_set ("DEFPSIZE");
    ssvn_system_set ("DEFUDFSVSIZ");
    ssvn_system_set ("DEFNSIZE");
    ssvn_system_set ("MAXNO_OF_RBUF");
    ssvn_system_set ("DEFNO_OF_RBUF");
    ssvn_system_set ("DEFPSIZE0");
    ssvn_system_set ("NO_GLOBLS");
    ssvn_system_set ("NO_OF_GBUF");
    ssvn_system_set ("NESTLEVLS");
    ssvn_system_set ("PARDEPTH");
    ssvn_system_set ("PATDEPTH");
    ssvn_system_set ("TRLIM");
    ssvn_system_set ("ARGS_IN_ESC");
    ssvn_system_set ("ZTLEN");
    ssvn_system_set ("FUNLEN");
    ssvn_system_set ("STRING_MAX");
    ssvn_system_set ("$NEXTOK");
    ssvn_system_set ("EOK");
    ssvn_system_set ("OFFOK");
    ssvn_system_set ("BIG_ENDIAN");
    ssvn_system_set ("ZDATE_FORMAT");
    ssvn_system_set ("ZTIME_FORMAT");
    
    fp = fopen (config_file, "r");

    while (fgets (line, 254, fp) != NULL) {

        if (line[0] == '[') {
            strcpy (line_ns, line + 1);
            line_ns[strlen (line_ns) - 2] = '\0';

            snprintf (k_buf, 254, "^$SYSTEM\202NAMESPACE\202%s\201", line_ns);
            symtab_shm (set_sym, k_buf, "\201");
        }
        
    }

    fclose (fp);

    snprintf (k_buf, 254, "^$SYSTEM\202CHANNELS\202TERMINAL\201");
    snprintf (d_buf, 254, "0,0\201");
    symtab_shm (set_sym, k_buf, d_buf);

    snprintf (k_buf, 254, "^$SYSTEM\202CHANNELS\202FILE\201");
    snprintf (d_buf, 254, "1,%d\201", FIRSTSCK - 1);
    symtab_shm (set_sym, k_buf, d_buf);

    snprintf (k_buf, 254, "^$SYSTEM\202CHANNELS\202SOCKET\201");
    snprintf (d_buf, 254, "%d,%d\201", FIRSTSCK, MAXDEV - 1);
    symtab_shm (set_sym, k_buf, d_buf);

    ssvn_job_update ();
    
}

void ssvn_system(short action, char *key, char *data)
{

    freem_ref_t *r = (freem_ref_t *) malloc (sizeof (freem_ref_t));
    NULLPTRCHK(r,"ssvn_system");  

    mref_init (r, MREF_RT_SSV, "^$SYSTEM");
    internal_to_mref (r, key);

    if ((r->subscript_count > 2) && (strcmp (r->subscripts[0], "EVENT") != 0) && (strcmp (r->subscripts[0], "MAPPINGS") != 0)) {
        merr_raise (INVREF);
        goto done;
    }

    switch (action) {
        
        case get_sym:

            if (strcmp (r->subscripts[0], "NAMESPACE") == 0) {

                symtab_shm (action, key, data);
                merr_clear ();
                break;
                
            }
            else if (strcmp (r->subscripts[0], "MAPPINGS") == 0) {

                if ((strcmp (r->subscripts[1], "GLOBAL") == 0) || (strcmp (r->subscripts[1], "ROUTINE") == 0)) {

                    merr_clear ();
                    symtab_shm (action, key, data);

                    break;

                }
                else {

                    merr_raise (M38);
                    break;

                }

            }
            else if (strcmp (r->subscripts[0], "ENVIRONMENT") == 0) {
                char varnam[255];
                char res[255];
                char *e;
                
                strcpy (varnam, r->subscripts[1]);

                e = getenv (varnam);

                if (e == NULL) {
                    sprintf (data, "\201");
                    break;
                }
                else {
                    strcpy (res, e);
                    stcnv_c2m (res);
                    stcpy (data, res);
                }

                break;
            }
            else if (strcmp (r->subscripts[0], "MAINTENANCE_MODE") == 0) {

                snprintf (data, 511, "%d\201", shm_config->hdr->maintenance_mode);
                merr_clear ();
                break;
                
            }
            else if (strcmp (r->subscripts[0], "DAEMON_PID") == 0) {

                snprintf (data, 511, "%d\201", shm_config->hdr->first_process);
                merr_clear ();
                break;
                
            }
            else if (strcmp (r->subscripts[0], "MWAPI") == 0) {
                
#if defined(MWAPI_GTK)
                snprintf (data, 511, "1\201");
#else
                snprintf (data, 511, "0\201");
#endif
                merr_clear ();
                break;
                
            }
            else if (strcmp (r->subscripts[0], "EVENT") == 0) {

                symtab_shm (action, key, data);
                merr_clear ();
                break;
                
            }
            else if (strcmp (r->subscripts[0], "ERRMSG") == 0) {

                int errnum;

                errnum = merr_code_to_num (r->subscripts[1]);

                if (errnum == -1) {
                    merr_raise (M38);
                    break;
                }
                
                sprintf (data, "%s", errmes[errnum]);
                merr_clear ();
                break;
                
            }
            else if (strcmp (r->subscripts[0], "NAME_LENGTH") == 0) {

                sprintf (data, "255\201");
                merr_clear ();
                break;
                
            }
            else if (strcmp (r->subscripts[0], "DEFPSIZE") == 0) {
                sprintf (data, "%ld\201", DEFPSIZE);
                merr_clear ();
                break;
            }
            else if (strcmp (r->subscripts[0], "DEFUDFSVSIZ") == 0) {
                sprintf (data, "%ld\201", (long) DEFUDFSVSIZ);
                merr_clear ();
                break;
            }
            else if (strcmp (r->subscripts[0], "DEFNSIZE") == 0) {
                sprintf (data, "%ld\201", (long) DEFNSIZE);
                merr_clear ();
                break;
            }
            else if (strcmp (r->subscripts[0], "MAXNO_OF_RBUF") == 0) {
                sprintf (data, "%d\201", MAXNO_OF_RBUF);
                merr_clear ();
                break;
            }
            else if (strcmp (r->subscripts[0], "DEFNO_OF_RBUF") == 0) {
                sprintf (data, "%ld\201", (long) DEFNO_OF_RBUF);
                merr_clear ();
                break;                
            }
            else if (strcmp (r->subscripts[0], "DEFPSIZE0") == 0) {
                sprintf (data, "%d\201", DEFPSIZE0);
                merr_clear ();
                break;
            }
            else if (strcmp (r->subscripts[0], "NO_GLOBLS") == 0) {
                sprintf (data, "%d\201", NO_GLOBLS);
                merr_clear ();
                break;
            }
            else if (strcmp (r->subscripts[0], "NO_OF_GBUF") == 0) {
                sprintf (data, "%d\201", NO_OF_GBUF);
                merr_clear ();
                break;
            }
            else if (strcmp (r->subscripts[0], "NESTLEVLS") == 0) {
                sprintf (data, "%d\201", NESTLEVLS);
                merr_clear ();
                break;
            }
            else if (strcmp (r->subscripts[0], "PARDEPTH") == 0) {
                sprintf (data, "%d\201", PARDEPTH);
                merr_clear ();
                break;
            }
            else if (strcmp (r->subscripts[0], "PATDEPTH") == 0) {
                sprintf (data, "%d\201", PATDEPTH);
                merr_clear ();
                break;
            }
            else if (strcmp (r->subscripts[0], "TRLIM") == 0) {
                sprintf (data, "%d\201", TRLIM);
                merr_clear ();
                break;
            }
            else if (strcmp (r->subscripts[0], "ARGS_IN_ESC") == 0) {
                sprintf (data, "%d\201", ARGS_IN_ESC);
                merr_clear ();
                break;
            }
            else if (strcmp (r->subscripts[0], "ZTLEN") == 0) {
                sprintf (data, "%d\201", ZTLEN);
                merr_clear ();
                break;
            }
            else if (strcmp (r->subscripts[0], "FUNLEN") == 0) {
                sprintf (data, "%d\201", FUNLEN);
                merr_clear ();
                break;
            }
            else if (strcmp (r->subscripts[0], "STRING_MAX") == 0) {

                sprintf (data, "%d\201", STRLEN);
                merr_clear ();
                break;

            }
            else if (strcmp (r->subscripts[0], "$NEXTOK") == 0) {

                sprintf (data, "1\201");
                merr_clear ();
                break;

            }
            else if (strcmp (r->subscripts[0], "EOK") == 0) {

                sprintf (data, "1\201");
                merr_clear ();
                break;

            }
            else if (strcmp (r->subscripts[0], "OFFOK") == 0) {

                sprintf (data, "1\201");
                merr_clear ();
                break;

            }
            else if (strcmp (r->subscripts[0], "BIG_ENDIAN") == 0) {

#if defined(HAVE_STDINT_H)                
                volatile uint32_t i = 0x01234567;

                if ((*((uint8_t*)(&i)) == 0x67) == 0) {
                    sprintf (data, "1\201");
                }
                else {
                    sprintf (data, "0\201");
                }
                
                merr_clear ();
                break;
#else
                merr_raise (M29);
                goto done;
#endif
            }
            else if (strcmp (r->subscripts[0], "ZDATE_FORMAT") == 0) {
                get_conf ("SYSTEM", "zdate_format", data);
                stcnv_c2m (data);
                
                
                break;
            }
            else if (strcmp (r->subscripts[0], "ZTIME_FORMAT") == 0) {
                get_conf ("SYSTEM", "ztime_format", data);
                stcnv_c2m (data);
                
                break;
            }
            else if (strcmp (r->subscripts[0], "CHANNELS") == 0) {
                symtab_shm (action, key, data);
                break;
            }
/*
            else {
                symtab_shm (action, key, data);
                break;
            }
*/            
            
            merr_raise (M29);
            goto done;

        case bigquery:
        case fra_query:
        case fra_order:
            /*
            if (strcmp (r->subscripts[0], "NAMESPACE") != 0) {
                merr_raise (M29);
                goto done;
            }
            */

            symtab_shm (action, key, data);
            merr_clear ();
            goto done;

        case kill_sym:
        case set_sym:
            if ((strcmp (r->subscripts[0], "MAINTENANCE_MODE") == 0) && (action == set_sym)) {
                shm_config->hdr->maintenance_mode = tvexpr (data);
                merr_clear ();
                goto done;
            }            
            
            if ((strcmp (r->subscripts[0], "EVENT") != 0) && (strcmp (r->subscripts[0], "MAPPINGS") != 0)) {
                merr_raise (M29);
                goto done;
            }

            if ((strcmp (r->subscripts[0], "MAPPINGS") == 0) && ((strcmp (r->subscripts[1], "GLOBAL") != 0) && (strcmp (r->subscripts[1], "ROUTINE") != 0))) {
                merr_raise (M38);
                goto done;
            }
            
            symtab_shm (action, key, data);
            merr_clear ();
            goto done;

        case fra_dat:
            symtab_shm (action, key, data);
            merr_clear ();
            goto done;
            
        default:
            merr_raise (INVREF);
            break;

    }

done:

    free (r);
    
    return;
}
